/**
 * Copyright (c) 2010-2016, Grill Balázs, IncQuery Labs Ltd.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * Contributors:
 * Grill Balázs - initial API and implementation
 */
package org.eclipse.viatra.query.runtime.localsearch.planner.cost.impl;

import com.google.common.base.Objects;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.viatra.query.runtime.localsearch.planner.cost.IConstraintEvaluationContext;
import org.eclipse.viatra.query.runtime.localsearch.planner.cost.ICostFunction;
import org.eclipse.viatra.query.runtime.matchers.context.IInputKey;
import org.eclipse.viatra.query.runtime.matchers.context.IQueryMetaContext;
import org.eclipse.viatra.query.runtime.matchers.context.IQueryRuntimeContext;
import org.eclipse.viatra.query.runtime.matchers.context.InputKeyImplication;
import org.eclipse.viatra.query.runtime.matchers.psystem.PConstraint;
import org.eclipse.viatra.query.runtime.matchers.psystem.PVariable;
import org.eclipse.viatra.query.runtime.matchers.psystem.basicdeferred.ExportedParameter;
import org.eclipse.viatra.query.runtime.matchers.psystem.basicdeferred.TypeFilterConstraint;
import org.eclipse.viatra.query.runtime.matchers.psystem.basicenumerables.ConstantValue;
import org.eclipse.viatra.query.runtime.matchers.psystem.basicenumerables.TypeConstraint;
import org.eclipse.viatra.query.runtime.matchers.tuple.Tuple;
import org.eclipse.xtend2.lib.StringConcatenation;

/**
 * Cost function which calculates cost based on the cardinality of items in the runtime model
 * 
 * @author Grill Balázs
 * @since 1.4
 */
@SuppressWarnings("all")
public abstract class StatisticsBasedConstraintCostFunction implements ICostFunction {
  private static double MAX_COST = 250.0;
  
  private static double DEFAULT_COST = (StatisticsBasedConstraintCostFunction.MAX_COST - 100.0);
  
  public abstract long countTuples(final IConstraintEvaluationContext input, final IInputKey supplierKey);
  
  @Override
  public double apply(final IConstraintEvaluationContext input) {
    PConstraint _constraint = input.getConstraint();
    return this.calculateCost(_constraint, input);
  }
  
  protected double _calculateCost(final ConstantValue constant, final IConstraintEvaluationContext input) {
    return 0.0f;
  }
  
  protected double _calculateCost(final TypeConstraint constraint, final IConstraintEvaluationContext input) {
    final Collection<PVariable> freeMaskVariables = input.getFreeVariables();
    final Collection<PVariable> boundMaskVariables = input.getBoundVariables();
    IInputKey supplierKey = ((TypeConstraint) constraint).getSupplierKey();
    long arity = supplierKey.getArity();
    if ((arity == 1)) {
      return this.calculateUnaryConstraintCost(constraint, input);
    } else {
      if ((arity == 2)) {
        long edgeCount = this.countTuples(input, supplierKey);
        Tuple _variablesTuple = ((TypeConstraint) constraint).getVariablesTuple();
        Object _get = _variablesTuple.get(0);
        PVariable srcVariable = ((PVariable) _get);
        Tuple _variablesTuple_1 = ((TypeConstraint) constraint).getVariablesTuple();
        Object _get_1 = _variablesTuple_1.get(1);
        PVariable dstVariable = ((PVariable) _get_1);
        boolean isInverse = false;
        if ((freeMaskVariables.contains(srcVariable) && boundMaskVariables.contains(dstVariable))) {
          isInverse = true;
        }
        return this.calculateBinaryExtendCost(supplierKey, srcVariable, dstVariable, isInverse, edgeCount, input);
      } else {
        StringConcatenation _builder = new StringConcatenation();
        _builder.append("Cost calculation for arity ");
        _builder.append(arity, "");
        _builder.append(" is not implemented yet");
        throw new RuntimeException(_builder.toString());
      }
    }
  }
  
  protected double calculateBinaryExtendCost(final IInputKey supplierKey, final PVariable srcVariable, final PVariable dstVariable, final boolean isInverse, final long edgeCount, final IConstraintEvaluationContext input) {
    final Collection<PVariable> freeMaskVariables = input.getFreeVariables();
    final Collection<PVariable> boundMaskVariables = input.getBoundVariables();
    final PConstraint constraint = input.getConstraint();
    IQueryRuntimeContext _runtimeContext = input.getRuntimeContext();
    IQueryMetaContext metaContext = _runtimeContext.getMetaContext();
    Collection<InputKeyImplication> implications = metaContext.getImplications(supplierKey);
    double srcCount = (-1);
    double dstCount = (-1);
    for (final InputKeyImplication implication : implications) {
      {
        List<Integer> impliedIndices = implication.getImpliedIndices();
        if (((impliedIndices.size() == 1) && impliedIndices.contains(Integer.valueOf(0)))) {
          IInputKey _impliedKey = implication.getImpliedKey();
          long _countTuples = this.countTuples(input, _impliedKey);
          srcCount = _countTuples;
        } else {
          if (((impliedIndices.size() == 1) && impliedIndices.contains(Integer.valueOf(1)))) {
            IInputKey _impliedKey_1 = implication.getImpliedKey();
            long _countTuples_1 = this.countTuples(input, _impliedKey_1);
            dstCount = _countTuples_1;
          }
        }
      }
    }
    if ((freeMaskVariables.contains(srcVariable) && freeMaskVariables.contains(dstVariable))) {
      return (dstCount * srcCount);
    } else {
      double srcNodeCount = (-1);
      double dstNodeCount = (-1);
      if (isInverse) {
        srcNodeCount = dstCount;
        dstNodeCount = srcCount;
      } else {
        srcNodeCount = srcCount;
        dstNodeCount = dstCount;
      }
      if (((srcNodeCount > (-1)) && (edgeCount > (-1)))) {
        if ((srcNodeCount == 0)) {
          return 0;
        } else {
          return (((double) edgeCount) / srcNodeCount);
        }
      } else {
        if (((srcCount > (-1)) && (dstCount > (-1)))) {
          if ((srcCount != 0)) {
            return (dstNodeCount / srcNodeCount);
          } else {
            return 1.0f;
          }
        } else {
          Map<Set<PVariable>, Set<PVariable>> functionalDependencies = constraint.getFunctionalDependencies(metaContext);
          Set<PVariable> impliedVariables = functionalDependencies.get(boundMaskVariables);
          if (((!Objects.equal(impliedVariables, null)) && impliedVariables.containsAll(freeMaskVariables))) {
            return 1.0;
          } else {
            return StatisticsBasedConstraintCostFunction.DEFAULT_COST;
          }
        }
      }
    }
  }
  
  protected double calculateUnaryConstraintCost(final TypeConstraint constraint, final IConstraintEvaluationContext input) {
    Tuple _variablesTuple = constraint.getVariablesTuple();
    Object _get = _variablesTuple.get(0);
    PVariable variable = ((PVariable) _get);
    Collection<PVariable> _boundVariables = input.getBoundVariables();
    boolean _contains = _boundVariables.contains(variable);
    if (_contains) {
      return 0.9;
    } else {
      IInputKey _supplierKey = constraint.getSupplierKey();
      long _countTuples = this.countTuples(input, _supplierKey);
      return (_countTuples + StatisticsBasedConstraintCostFunction.DEFAULT_COST);
    }
  }
  
  protected double _calculateCost(final ExportedParameter exportedParam, final IConstraintEvaluationContext input) {
    return 0.0;
  }
  
  protected double _calculateCost(final TypeFilterConstraint exportedParam, final IConstraintEvaluationContext input) {
    return 0.0;
  }
  
  /**
   * Default cost calculation strategy
   */
  protected double _calculateCost(final PConstraint constraint, final IConstraintEvaluationContext input) {
    Collection<PVariable> _freeVariables = input.getFreeVariables();
    boolean _isEmpty = _freeVariables.isEmpty();
    if (_isEmpty) {
      return 1.0;
    } else {
      return StatisticsBasedConstraintCostFunction.DEFAULT_COST;
    }
  }
  
  public double calculateCost(final PConstraint exportedParam, final IConstraintEvaluationContext input) {
    if (exportedParam instanceof ExportedParameter) {
      return _calculateCost((ExportedParameter)exportedParam, input);
    } else if (exportedParam instanceof TypeFilterConstraint) {
      return _calculateCost((TypeFilterConstraint)exportedParam, input);
    } else if (exportedParam instanceof ConstantValue) {
      return _calculateCost((ConstantValue)exportedParam, input);
    } else if (exportedParam instanceof TypeConstraint) {
      return _calculateCost((TypeConstraint)exportedParam, input);
    } else if (exportedParam != null) {
      return _calculateCost(exportedParam, input);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(exportedParam, input).toString());
    }
  }
}
